'''
This software is released under an MIT/X11 open source license.

Copyright 2013-2014 Diffeo, Inc.
'''
from __future__ import absolute_import
import collections
import contextlib
import importlib
import logging
import pdb
import os

import yaml

logger = logging.getLogger('yakonfig')

_runtime_args_object = None
_runtime_args_dict = None

_config_file_path = None

_config_cache = None


def set_runtime_args_object(args):
    """Set an agrguments object, from argparse ArgumentParser.parse_args() or similar. Will be used for substituting into !runtime values in the config yaml."""
    global _runtime_args_object
    global _runtime_args_dict
    _runtime_args_object = args
    _runtime_args_dict = None


def set_runtime_args_dict(args):
    """Set a dictionary of global options. Will be used for substituting into !runtime values in the config yaml."""
    global _runtime_args_object
    global _runtime_args_dict
    if args:
        _runtime_args_dict = args
        _runtime_args_object = None


class Loader(yaml.Loader):

    def __init__(self, stream):
        ## find root path for !include relative path
        streamname = getattr(stream, 'name', None)
        if streamname:
            self._root = os.path.dirname(streamname)
        else:
            self._root = None
        super(Loader, self).__init__(stream)

    def include_func(self, node):
        '''
        call a python function to inject values into the yaml, such as
        a function that provides default yaml for loading.
        '''
        mod_name = self.construct_scalar(node)
        parts = mod_name.split('.')
        if not len(parts) >= 2:
            raise Exception('!include_func expects full.path.to.func(), not %r' % mod_name)
        func_name = parts[-1]
        mod_name = '.'.join(parts[:-1])
        mod = importlib.import_module(mod_name)
        func = getattr(mod, func_name, None)
        if not func:
            raise Exception('%r not found in %r, dir(%r) = %r' % (func_name, mod_name, mod_name, dir(mod)))

        if func_name.endswith('yaml'):
            ## functions named ".*yaml$" must return YAML to which we
            ## apply this Loader
            return yaml.load(func(), Loader)
        else:
            return func()

    def include_yaml(self, node):
        '''
        load another yaml file from the path specified by node's value
        '''
        filename = self.construct_scalar(node)
        if not filename.startswith('/'):
            if self._root is None:
                raise Exception('!include_yaml %s is a relative path, but stream lacks path' % filename)
            filename = os.path.join(self._root, self.construct_scalar(node))
        with self.open(filename, 'r') as fin:
            return yaml.load(fin, Loader)

    def include_runtime(self, node):
        '''
        load another yaml file from the path specified by runtime arg
        named by node's value
        '''
        if _runtime_args_dict is None and _runtime_args_object is None:
            raise Exception('!runtime requires a prior call to set_runtime_args_dict or set_runtime_args_object')
        runtimedict = _runtime_args_dict or vars(_runtime_args_object)
        filename = runtimedict.get(node.value)
        if filename is None:
            raise Exception('%r not in runtime args: %r' % (node.value, runtimedict))
        if not filename.startswith('/'):
            if self._root is None:
                raise Exception('!include_runtime %s is a relative path, but stream lacks path' % filename)
            filename = os.path.join(self._root, self.construct_scalar(node))
        with self.open(filename, 'r') as fin:
            return yaml.load(fin, Loader)

    def open(self, *args, **kwargs):
        '''
        method that looks like the regular python builtin `open`, and
        an be replaced by tests with different behavior
        '''
        return open(*args, **kwargs)

    def runtime(self, node):
        '''
        provide !runtime values from having set_runtime_args_dict or
        set_runtime_args_object
        '''
        if _runtime_args_dict is None and _runtime_args_object is None:
            raise Exception('!runtime requires a prior call to set_runtime_args_dict or set_runtime_args_object')
        runtimedict = _runtime_args_dict or vars(_runtime_args_object)
        if (node is None) or (not node.value):
            return runtimedict  # with no specifier, return the whole thing
        return runtimedict.get(node.value)


Loader.add_constructor('!include_runtime', Loader.include_runtime)
Loader.add_constructor('!include_func', Loader.include_func)
Loader.add_constructor('!include_yaml', Loader.include_yaml)
Loader.add_constructor('!runtime', Loader.runtime)


def clear_global_config():
    global _config_cache, _config_file_path, _runtime_args_object, _runtime_args_dict
    _config_cache = None
    _runtime_args_object = None
    _runtime_args_dict = None
    _config_file_path = None


def set_global_config(path_dict_or_stream):
    """Usage: call this from main() with a file system path, stream
    object, or a dict.  Calling it repeatedly with the same path is
    safe.  Calling it with a different path or repeatedly with a
    stream or dict requires an explicit call to clear_global_config.
    """
    path = None
    mapping = None
    stream = None

    global _config_file_path
    global _config_cache

    if isinstance(path_dict_or_stream, basestring):
        path = path_dict_or_stream
        if _config_file_path and _config_file_path != path:
            raise Exception('set_global_config(%r) differs from %r, consider calling clear_global_config first' % (path, _config_file_path))
        _config_file_path = path
        stream = open(path)

    elif isinstance(path_dict_or_stream, collections.Mapping):
        mapping = path_dict_or_stream

    elif hasattr(path_dict_or_stream, 'read'):
        stream = path_dict_or_stream

    else:
        raise Exception('set_global_config(%r) instead of a path, mapping object, or stream open for reading' % path_dict_or_stream)

    if stream is not None:
        mapping = yaml.load(stream, Loader)

    if _config_cache is not None:
        logger.warn('resetting config to all new values')

    _config_cache = mapping

    # TODO: convert to frozen dict?
    return _config_cache

def get_global_config(*args):
    """Get (a subset of) the global configuration.

    If no arguments are provided, returns the entire configuration.
    Otherwise, start with the entire configuration, and get the item
    named by the first parameter; then search that for the second
    parameter; and so on.

    :param args: configuration name path to fetch
    :return: configuration item or subtree
    :raise KeyError: if an argument is missing

    """
    global _config_cache
    c = _config_cache
    if c is None:
        if len(args) == 0: args = (None,)
        raise KeyError(args[0])
    for a in args:
        c = c[a]
    return c

@contextlib.contextmanager
def _temporary_config():
    """Temporarily replace the global configuration.

    Use this in a 'with' statement.  The inner block may freely manipulate
    the global configuration; the original global configuration is restored
    at exit.

    """
    global _config_cache, _config_file_path, _runtime_args_object, _runtime_args_dict
    old_cc = _config_cache
    old_cfp = _config_file_path
    old_rao = _runtime_args_object
    old_rad = _runtime_args_dict
    clear_global_config()
    yield
    _config_cache = old_cc
    _config_file_path = old_cfp
    _runtime_args_object = old_rao
    _runtime_args_dict = old_rad



"""Top-level entry points to yakonfig.

Purpose
=======

Most programs' `main()` functions will call yakonfig as:

>>> parser = argparse.ArgumentParser()
>>> yakonfig.parse_args(parser, [yakonfig, module, module...])

where the list of modules are top-level modules or other
`yakonfig.Configurable` objects the program uses.

Test code and other things not driven by argparse can instead call

>>> yakonfig.set_default_config([yakonfig, module, module, ...])


Implementation Details
======================


-----

This software is released under an MIT/X11 open source license.

Copyright 2014 Diffeo, Inc.

"""

from __future__ import absolute_import
import collections
import contextlib
from cStringIO import StringIO

import yaml

import yakonfig
from yakonfig.merge import overlay_config
from yakonfig.yakonfig import _temporary_config

# These implement the Configurable interface for yakonfig proper!
config_name = 'yakonfig'
def add_arguments(parser):
    parser.add_argument('--config', '-c', metavar='FILE',
                        help='read configuration from FILE')
runtime_keys = { 'config': 'config' }

def parse_args(parser, modules, args=None):
    """Set up global configuration for command-line tools.

    `modules` is an iterable of `yakonfig.Configurable` objects, or
    anything equivalently typed.  This function iterates through those
    objects and calls `add_arguments()` on each to build up a complete
    list of command-line arguments, then calls
    `argparse.ArgumentParser.parse_args()` to actually process the
    command line.  This produces a configuration that is a combination
    of all default values declared by all modules; configuration
    specified in `--config` arguments; and overriding configuration
    values specified in command-line arguments.

    This returns the ArgumentParser Namespace object, in case the
    application has defined its own command-line parameters and
    needs to process them.  The new global configuration can be
    obtained via `yakonfig.get_global_config()`.

    :param argparse.ArgumentParser parser: application-provided
      argument parser
    :param iterable modules: modules or Configurable instances to use
    :param args: command-line options, or `None` to use `sys.argv`
    :return: the new global configuration

    """
    collect_add_argparse(parser, modules)
    namespace = parser.parse_args(args)
    try:
        yakonfig.set_default_config(modules, params=vars(namespace))
    except yakonfig.ConfigurationError, e:
        parser.error(e)
    return namespace

def set_default_config(modules, params={}, yaml=None, filename=None,
                       config=None, validate=True):
    """Set up global configuration for tests and noninteractive tools.

    `modules` is an iterable of `yakonfig.Configurable' objects, or
    anything equivalently typed.  This function iterates through those
    objects to produce a default configuration, reads `yaml` as though
    it were the configuration file, and fills in any values from
    `params` as though they were command-line arguments.  The
    resulting configuration is set as the global configuration.

    :param iterable modules: modules or Configurable instances to use
    :param dict params: dictionary of command-line argument key to values
    :param str yaml: global configuration file
    :param str filename: location of global configuration file
    :param dict config: global configuration object
    :param bool validate: check configuration after creating
    :return: the new global configuration

    """

    # Get the configuration from the file, or from params['config']
    file_config = {}
    if yaml is None and filename is None and config is None:
        if 'config' in params and params['config'] is not None:
            filename = params['config']
    if yaml is not None or filename is not None or config is not None:
        import yaml as y
        if yaml is not None:
            file_config = y.load(StringIO(yaml))
        elif filename is not None:
            with open(filename, 'r') as f:
                file_config = y.load(f)
        elif config is not None:
            file_config = config

    # Assemble the configuration from defaults + file + arguments
    default_config = assemble_default_config(modules)
    base_config = overlay_config(default_config, file_config)
    fill_in_arguments(base_config, modules, params)
    
    # Replace the modules list (accommodate external modules)
    def replace_module(config, m):
        name = getattr(m, 'config_name')
        c = config.get(name, {})
        if hasattr(m, 'replace_config'):
            return getattr(m, 'replace_config')(c, name)
        return m
    modules = [replace_module(base_config, m) for m in modules]

    # Reassemble the configuration again
    default_config = assemble_default_config(modules)
    base_config = overlay_config(default_config, file_config)
    fill_in_arguments(base_config, modules, params)

    # Validate the configuration
    if validate and len(modules) > 0:
        mod = modules[-1]
        checker = getattr(mod, 'check_config', None)
        if checker is not None:
            with _temporary_config():
                yakonfig.set_global_config(base_config)
                checker(base_config[mod.config_name], mod.config_name)

    # All done, set the global configuration
    yakonfig.set_global_config(base_config)
    return base_config

@contextlib.contextmanager
def defaulted_config(modules, params={}, yaml=None, filename=None,
                     config=None, validate=True):
    """Context manager version of `set_default_config()`.

    Use this with a Python 'with' statement, like

    >>> config_yaml = '''
    ... toplevel:
    ...   param: value
    ... '''
    >>> with yakonfig.defaulted_config([toplevel], yaml=config_yaml) as config:
    ...    assert 'param' in config['toplevel']
    ...    assert yakonfig.get_global_config('toplevel', 'param') == 'value'

    On exit the global configuration is restored to its previous state
    (if any).

    :param iterable modules: modules or Configurable instances to use
    :param dict params: dictionary of command-line argument key to values
    :param str yaml: global configuration file
    :param str filename: location of global configuration file
    :param dict config: global configuration object
    :param bool validate: check configuration after creating
    :return: the new global configuration

    """
    with _temporary_config():
        set_default_config(modules, params=params, yaml=yaml,
                           filename=filename, config=config, validate=validate)
        yield yakonfig.get_global_config()

def check_toplevel_config(what, who):
    """Verify that some dependent configuration is present and correct.

    This will generally be called from a `check_config` implementation.
    `what` is a Configurable-like object.  If the corresponding
    configuration isn't present in the global configuration, raise a
    `ConfigurationError` explaining that `who` required it.  Otherwise
    call that module's `check_config` (if any).

    :param Configurable what: top-level module to require
    :param str who: name of the requiring module
    :raise ConfigurationError: if configuration for `what` is missing
      or incorrect

    """
    config_name = what.config_name
    config = yakonfig.get_global_config()
    if config_name not in config:
        raise yakonfig.ConfigurationError(
            '{} requires top-level configuration for {}'
            .format(who, config_name))
    checker = getattr(what, 'check_config', None)
    if checker:
        checker(config[config_name], config_name)

def _walk_config(config, modules, f, prefix='', required=True):
    """Recursively walk through a module list.

    For every module, calls ``f(config, module, name)`` where
    `config` is the configuration scoped to that module, `module`
    is the Configurable-like object, and `name` is the complete
    path (ending in the module name).

    If `required` is true, every name encountered must exist,
    and `yakonfig.ProgrammerError` will be raised if it doesn't.
    If `required` is false, every name encountered must not exist,
    and `yakonfig.ProgrammerError` will be raised if it does.
    `required=False` is intended to only be used when setting up
    the default configuration.

    :param dict config: configuration to walk and possibly update
    :param iterable modules: modules or Configurable instances to use
    :param f: callback function for each module
    :param str prefix: prefix name of the config
    :param bool required: if true, config names already exist
    :return: config

    """
    for module in modules:
        name = getattr(module, 'config_name', None)
        if name is None:
            raise yakonfig.ProgrammerError('{!r} must provide a config_name'
                                           .format(module))
        new_name = '{}{}'.format(prefix, name)
        if required:
            if name not in config:
                raise yakonfig.ProgrammerError('{} not present in configuration'
                                               .format(new_name))
            if not isinstance(config[name], collections.Mapping):
                raise yakonfig.ConfigurationError(
                    '{} must be an object configuration'.format(new_name))
        else:
            if name in config:
                raise yakonfig.ProgrammerError('multiple modules providing {}'
                                               .format(new_name))
            config[name] = {}

        # do the work!
        f(config[name], module, new_name)

        # recurse into submodules (if defined)
        _walk_config(config=config[name],
                     modules=getattr(module, 'sub_modules', []),
                     f=f,
                     prefix=new_name + '.',
                     required=required)
    return config

def collect_add_argparse(parser, modules):
    """Add all command-line options.

    `modules` is an iterable of `yakonfig.Configurable` objects, or
    anything equivalently typed.  This calls `add_arguments` (if
    present) on all of them to set the global command-line arguments.

    :param argparse.ArgumentParser parser: argparse parser
    :param iterable modules: modules or Configurable instances to use

    """
    def work_in(config, module, name):
        f = getattr(module, 'add_arguments', None)
        if f is not None:
            f(parser)
    # this is *really handy* so we'll use it -- but it builds up a
    # partial config tree that we'll 100% ignore
    _walk_config(dict(), modules, work_in, required=False)
    return parser

def assemble_default_config(modules):

    """Build the default configuration from a set of modules.

    `modules` is an iterable of `yakonfig.Configurable` objects, or
    anything equivalently typed.  This produces the default
    configuration from that list of modules.

    :param iterable modules: modules or Configurable instances to use
    :return: configuration dictionary

    """
    def work_in(config, module, name):
        local_config = dict(getattr(module, 'default_config', {}))
        config.update(local_config)
    return _walk_config(dict(), modules, work_in, required=False)

def fill_in_arguments(config, modules, args):
    """Fill in configuration fields from command-line arguments.

    `config` is a dictionary holding the initial configuration,
    probably the result of `assemble_default_config()`.  It reads
    through `modules`, and for each, fills in any configuration values
    that are provided in `args`.

    `config` is modified in place.  `args` may be either a dictionary
    or an object (as the result of argparse).

    :param dict config: configuration tree to update
    :param iterable modules: modules or Configurable instances to use
    :param args: command-line objects
    :paramtype args: dict or object
    :return: config

    """
    def work_in(config, module, name):
        rkeys = getattr(module, 'runtime_keys', {})
        for (attr, cname) in rkeys.iteritems():
            v = args.get(attr, None)
            if v is not None:
                config[cname] = v
    if not isinstance(args, collections.Mapping):
        args = vars(args)
    return _walk_config(config, modules, work_in)



'''
This software is released under an MIT/X11 open source license.

Copyright 2013-2014 Diffeo, Inc.


Loads a yaml file (with extensions) and makes it available globally to a Python application.


YAML extensions:

key: !runtime [rkey]

From some runtime set of options (via argparse or a dict of options) insert that value under key. If 'rkey' is specified then do dict or property access by that name and use that value instead of the whole. [See set_runtime_args_dict(args)]


key: !include_yaml path

Loads a yaml file at path and inserts it as the value associated with key.


key: !include_func package.path.to.func

Calls a python function. from a fully specified name of package.func
If the function name ends in "yaml" the return value is interpreted as a yaml document body in a string and parsed. Otherwise the return value is assumed to be a dict or other object that can be simply assigned to they key at this point in the yaml file.


key: !include_runtime rkey

Like !runtime pulls a value from input to set_runtime_args_dict(), but uses that value as a path as in !include_yaml, reading that file and parsing it and inserting it at this point in the enclosing yaml file.



get_global_config(path=None, stream=None)

Call this once from main() with a path or stream of config yaml to parse.
Call this throughout the application without any arguments to get the cached global config dictionary.


set_runtime_args_object(args)

Set an object to reference in !runtime directives in the config yaml.
e.g. ap = argparse.ArgumentParser()
ap.add_argument('--foo')
set_runtime_args_object(ap.parse_args())


set_runtime_args_dict(args)

Set a dictionary to be reference by !runtime directives in the config yaml.
'''
from __future__ import absolute_import
from yakonfig.configurable import Configurable, ProxyConfigurable, \
    NewSubModules, check_subconfig
from yakonfig.exceptions import *
from yakonfig.merge import diff_config, overlay_config
from yakonfig.toplevel import parse_args, set_default_config, \
    defaulted_config, check_toplevel_config, \
    config_name, add_arguments, runtime_keys
from yakonfig.yakonfig import set_runtime_args_object, set_runtime_args_dict, \
    clear_global_config, set_global_config, get_global_config
